package edu.mit.harold.icollections;

import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;




/**
 * 
 * An efficient immutable map from integer keys to non-null values.
 * <p>
 * Iteration occurs in the integer order of the keys.
 * <p>
 * The balanced tree is based on the Glasgow Haskell Compiler's Data.Map implementation,
 * which in turn is based on "size balanced binary trees" as described by:
 * <p>
 * Stephen Adams, "Efficient sets: a balancing act",
 * Journal of Functional Programming 3(4):553-562, October 1993,
 * http://www.swiss.ai.mit.edu/~adams/BB/.
 * <p>
 * J. Nievergelt and E.M. Reingold, "Binary search trees of bounded balance",
 * SIAM journal of computing 2(1), March 1973.
 * 
 * @author harold
 *
 * @param <V>
 */
public final class IntTreeIMap<V> extends AbstractMap<Integer,V> implements IMap<Integer,V> {
//// STATIC FACTORY METHODS ////
	private static final IntTreeIMap<Object> EMPTY = new IntTreeIMap<Object>(IntTree.EMPTYNODE);

	/**
	 * @param <V>
	 * @return an empty map
	 */
	@SuppressWarnings("unchecked")
	public static <V> IntTreeIMap<V> empty() {
		return (IntTreeIMap<V>)EMPTY; }
	
	/**
	 * @param <V>
	 * @param key
	 * @param value
	 * @return empty().plus(key, value)
	 */
	public static <V> IntTreeIMap<V> singleton(final Integer key, final V value) {
		return IntTreeIMap.<V>empty().plus(key, value); }
	
	/**
	 * @param <V>
	 * @param map
	 * @return empty().plusAll(map)
	 */
	@SuppressWarnings("unchecked")
	public static <V> IntTreeIMap<V> from(final Map<? extends Integer, ? extends V> map) {
		if(map instanceof IntTreeIMap)
			return (IntTreeIMap<V>)map; //(actually we only know it's IntTreeIMap<? extends V>)
									// but that's good enough for an immutable
									// (i.e. we can't mess someone else up by adding the wrong type to it)
		return IntTreeIMap.<V>empty().plusAll(map); }
	

//// PRIVATE CONSTRUCTORS ////
	private final IntTree<V> root;
	// not externally instantiable (or subclassable):
	private IntTreeIMap(final IntTree<V> root) {
		// TODO constructor should take hashCode?
		// would require extra code in plus*() and minus*(),
		// but probably won't impact performance at all,
		// except for huge improvement in hashCode().
		// yeah but who's gonna use this map anyway?
		this.root = root; }
	private IntTreeIMap<V> withRoot(final IntTree<V> root) {
		if(root==this.root) return this;
		return new IntTreeIMap<V>(root); }
	

//// UNINHERITED METHODS OF IntTreeImap ////
	IntTreeIMap<V> withKeysChangedAbove(final int key, final int delta) {
		// TODO check preconditions of changeKeysAbove()
		// TODO make public?
		return withRoot( root.changeKeysAbove(key, delta) );
	}
	
	IntTreeIMap<V> withKeysChangedBelow(final int key, final int delta) {
		// TODO check preconditions of changeKeysAbove()
		// TODO make public?
		return withRoot( root.changeKeysBelow(key, delta) );
	}
	
//// REQUIRED METHODS FROM AbstractMap ////
	private Set<Entry<Integer,V>> entrySet = null;
	@Override
	public Set<Entry<Integer,V>> entrySet() {
		if(entrySet==null)
			entrySet = new AbstractSet<Entry<Integer,V>>() {
				// REQUIRED METHODS OF AbstractSet // 
				@Override
				public int size() { // same as Map
					return IntTreeIMap.this.size(); }
				@Override
				public Iterator<Entry<Integer,V>> iterator() {
					return root.iterator(); }
				// OVERRIDDEN METHODS OF AbstractSet //
				@Override
				public boolean contains(final Object e) {
					if(!(e instanceof Entry))
						return false;
					V value = get(((Entry<?,?>)e).getKey());
					return value!=null && value.equals(((Entry<?,?>)e).getValue());
				}
				@Override
				public int hashCode() { // same as Map
					// this is not an infinite loop,
					// since AbstractMap.hashCode() uses entrySet().iterator(),
					// not entrySet().hashCode():
					return IntTreeIMap.this.hashCode(); }
			};
		return entrySet;
	}

	
//// OVERRIDDEN METHODS FROM AbstractMap ////
	private Integer hashCode = null;
	@Override
	public int hashCode() {
		if(hashCode==null)
			hashCode = super.hashCode(); // AbstractMap.hashCode(), uses entrySet()
		return hashCode;
	}
	
	@Override
	public int size() {
		return root.size(); }

	@Override
	public boolean containsKey(final Object key) {
		if(!(key instanceof Integer))
			return false;
		return root.containsKey((Integer)key);
	}
	
	@Override
	public V get(final Object key) {
		if(!(key instanceof Integer))
			return null;
		return root.get((Integer)key);
	}

	
//// IMPLEMENTED METHODS OF IMap////
	public IntTreeIMap<V> plus(final Integer key, final V value) {
		return withRoot( root.plus(key, value) ); }
	
	public IntTreeIMap<V> minus(final Object key) {
		if(!(key instanceof Integer)) return this;
		return withRoot( root.minus((Integer)key) ); }
	
	public IntTreeIMap<V> plusAll(final Map<? extends Integer, ? extends V> map) {
		IntTree<V> root = this.root;
		for(Entry<? extends Integer,? extends V> entry : map.entrySet())
			root = root.plus(entry.getKey(), entry.getValue());
		return withRoot(root);
	}

	public IntTreeIMap<V> minusAll(final Collection<?> keys) {
		IntTree<V> root = this.root;
		for(Object key : keys)
			if(key instanceof Integer)
				root = root.minus((Integer)key);
		return withRoot(root);
	}
}
